@@ -93,6 +93,10 @@ struct fb_fillrect {
 static int TTY = 0;
 
 static int fb_hgpm = -2;
+static int fb_hmice = -1;
+static unsigned char ps2_mouse_mode;
+
+#define fb_have_mouse	(fb_hgpm >= 0 || fb_hmice >= 0)
 
 static int fb_console;
 
@@ -208,6 +212,9 @@ static int last_mouse_buttons;
 	int mouse_hidden;\
 	int xs, ys;\
 	unsigned char *data=bmp->data;\
+\
+	if (!data)\
+		return;\
 \
 	TEST_INACTIVITY\
 	CLIP_DRAW_BITMAP\
@@ -320,8 +327,8 @@ flush_ret:
 
 
 
-#define mouse_getscansegment(buf,x,y,w) memcpy(buf, fb_vmem + y * fb_linesize + x * fb_pixelsize, w)
-#define mouse_drawscansegment(ptr,x,y,w) memcpy_to_fb(fb_vmem + y * fb_linesize + x * fb_pixelsize, ptr, w)
+#define mouse_getscansegment(buf,x,y,w) memcpy_to_fb(buf, fb_vmem + y * fb_linesize + x * fb_pixelsize, w, 1)
+#define mouse_drawscansegment(ptr,x,y,w) memcpy_to_fb(fb_vmem + y * fb_linesize + x * fb_pixelsize, ptr, w, 0)
 
 #define do_with_mouse_device(cmd)				\
 do {								\
@@ -411,6 +418,9 @@ static void place_mouse(void)
 {
 	struct bitmap bmp;
 
+	if (!fb_have_mouse)
+		return;
+
 	bmp.x=arrow_width;
 	bmp.y=arrow_height;
 	bmp.skip=arrow_width*fb_pixelsize;
@@ -539,8 +549,8 @@ static inline void place_mouse_composite(void)
 			mouse_drawscansegment(mouse_ptr,mouse_left,mouse_top,l1);
 			if (background_length>0)
 				mouse_drawscansegment(
-					background_ptr +l2,
-					mouse_right,mouse_top ,l3);
+					background_ptr + l2,
+					mouse_right,mouse_top, l3);
 			mouse_ptr+=skip;
 			background_ptr+=skip;
 		}
@@ -892,7 +902,7 @@ static unsigned char *fb_switch_init(void)
 static void fb_switch_shutdown(void)
 {
 	int rs;
-	if (in_gr_operation <= 0) internal("fb_switch_shutdown: in_gr_operation %d", in_gr_operation);
+	if (in_gr_operation <= 0) internal_error("fb_switch_shutdown: in_gr_operation %d", in_gr_operation);
 	if (!fb_active && in_gr_operation == 1) {
 		EINTRLOOP(rs, ioctl(TTY, VT_RELDISP, 1));
 	}
@@ -932,10 +942,13 @@ static void fb_gpm_in(void *nic)
 	restore_gpm_signals();
 	if (g <= 0) {
 		fb_hgpm = -1;
+		hide_mouse();
 		unhandle_fb_mouse();
 		return;
 	}
 	set_handlers(fb_hgpm, fb_gpm_in, (void (*)(void *))NULL, NULL);
+	if (fb_hmice != -1)
+		return;
 	/*fprintf(stderr, "%x %x %d %d %d %d\n", gev.type, gev.buttons, gev.dx, gev.dy, gev.wdx, gev.wdy);*/
 	if (gev.dx || gev.dy) {
 		if (!((int)gev.type & gpm_smooth)) {
@@ -1016,9 +1029,124 @@ static void fb_gpm_in(void *nic)
 	EINTRLOOP(rs, fsync(fb_handle));
 }
 
+static void fb_ps2_in(void *nic)
+{
+	unsigned char buffer[4];
+
+	int rs, dx, dy, dw, button;
+
+	int size = ps2_mouse_mode ? 4 : 3;
+	if (hard_read(fb_hmice, buffer, size) != size) {
+		unhandle_fb_mouse();
+		return;
+	}
+
+	dx = buffer[1];
+	if (buffer[0] & 0x10) dx |= -0x100;
+	dy = buffer[2];
+	if (buffer[0] & 0x20) dy |= -0x100;
+	dy = -dy;
+
+	button = -1;
+	if (buffer[0] & 1) button = B_LEFT;
+	else if (buffer[0] & 2) button = B_RIGHT;
+	else if (buffer[0] & 4) button = B_MIDDLE;
+	else if (ps2_mouse_mode == 4) {
+		if (buffer[3] & 0x10) button = B_FOURTH;
+		else if (buffer[3] & 0x20) button = B_FIFTH;
+	}
+
+	dw = 0;
+	if (ps2_mouse_mode == 3) {
+		dw = buffer[3];
+		if (dw & 0x80) dw |= -0x100;
+	}
+	if (ps2_mouse_mode == 4) {
+		dw = buffer[3] & 0x0f;
+		if (dw & 0x08) dw |= -0x10;
+	}
+
+	mouse_x += dx;
+	mouse_y += dy;
+
+	if (mouse_x >= fb_xsize) mouse_x = fb_xsize - 1;
+	if (mouse_y >= fb_ysize) mouse_y = fb_ysize - 1;
+	if (mouse_x < 0) mouse_x = 0;
+	if (mouse_y < 0) mouse_y = 0;
+
+	if ((last_mouse_buttons & BM_ACT) == B_UP || (last_mouse_buttons & BM_ACT) == B_MOVE) {
+		if (button == -1)
+			button = B_MOVE;
+		else
+			button |= B_DOWN;
+	} else {
+		if (button == -1)
+			button = (last_mouse_buttons & BM_BUTT) | B_UP;
+		else
+			button |= B_DRAG;
+	}
+
+	if ((button & BM_ACT) == B_DRAG || (button & BM_ACT) == B_MOVE) {
+		if (dw > 0) {
+			button &= ~BM_BUTT;
+			button |= B_WHEELDOWN;
+		} else if (dw < 0) {
+			button &= ~BM_BUTT;
+			button |= B_WHEELUP;
+		}
+	}
+
+	/*fprintf(stderr, "%02x %d %d %d %d\n", buffer[0], mouse_x, mouse_y, button, last_mouse_buttons);*/
+
+	last_mouse_buttons = button;
+	if (!current_virtual_device) return;
+	if (current_virtual_device->mouse_handler) current_virtual_device->mouse_handler(current_virtual_device, mouse_x, mouse_y, button);
+	redraw_mouse();
+	EINTRLOOP(rs, fsync(fb_handle));
+}
+
 static int handle_fb_mouse(void)
 {
 	Gpm_Connect conn;
+
+	fb_hmice = c_open(cast_uchar "/dev/input/mice", O_RDWR | O_NOCTTY);
+	if (fb_hmice != -1) {
+		unsigned char c;
+		static const unsigned char imps_seq[] = { 0xf3, 200, 0xf3, 100, 0xf3, 80, 0xf2 };
+		static const unsigned char imex_seq[] = { 0xf3, 200, 0xf3, 200, 0xf3, 80, 0xf2 };
+		if (hard_write(fb_hmice, imps_seq, sizeof(imps_seq)) != sizeof(imps_seq))
+			goto gpm_fallback;
+		do
+			if (hard_read(fb_hmice, &c, 1) != 1)
+				goto gpm_fallback;
+		while (c != 0xfa);
+		if (hard_read(fb_hmice, &ps2_mouse_mode, 1) != 1)
+			goto gpm_fallback;
+		if (!ps2_mouse_mode)
+			goto open_gpm;
+		if (ps2_mouse_mode != 3)
+			goto gpm_fallback;
+
+		if (hard_write(fb_hmice, imex_seq, sizeof(imex_seq)) != sizeof(imex_seq))
+			goto gpm_fallback;
+		do
+			if (hard_read(fb_hmice, &c, 1) != 1)
+				goto gpm_fallback;
+		while (c != 0xfa);
+		if (hard_read(fb_hmice, &ps2_mouse_mode, 1) != 1)
+			goto gpm_fallback;
+		if (ps2_mouse_mode == 3)
+			goto open_gpm;
+		if (ps2_mouse_mode != 4)
+			goto gpm_fallback;
+	}
+
+open_gpm:
+	if (0) {
+gpm_fallback:
+		close_socket(&fb_hmice);
+	}
+
 	fb_hgpm = -2;
 	conn.eventMask = (unsigned short)~0U;
 	conn.defaultMask = gpm_smooth;
@@ -1029,10 +1157,16 @@ static int handle_fb_mouse(void)
 	restore_gpm_signals();
 	if (fb_hgpm < 0) {
 		fb_hgpm = -2;
+		if (fb_hmice >= 0)
+			goto done;
 		unhandle_fb_mouse();
 		return -1;
 	}
-	set_handlers(fb_hgpm, fb_gpm_in, (void (*)(void *))NULL, NULL);
+done:
+	if (fb_hgpm >= 0)
+		set_handlers(fb_hgpm, fb_gpm_in, (void (*)(void *))NULL, NULL);
+	if (fb_hmice != -1)
+		set_handlers(fb_hmice, fb_ps2_in, (void (*)(void *))NULL, NULL);
 #ifdef SIGTSTP
 	install_signal_handler(SIGTSTP, sig_tstp, NULL, 0);
 #endif
@@ -1055,6 +1189,8 @@ static void unhandle_fb_mouse(void)
 		restore_gpm_signals();
 	}
 	fb_hgpm = -2;
+
+	close_socket(&fb_hmice);
 #ifdef SIGTSTP
 	install_signal_handler(SIGTSTP, sig_tstp, NULL, 0);
 #endif
@@ -1098,6 +1234,7 @@ static void fb_pan_display(void)
 
 static unsigned char *fb_init_driver(unsigned char *param, unsigned char *ignore)
 {
+	char *se;
 	unsigned char *e;
 	struct stat st;
 	int rs;
@@ -1136,13 +1273,13 @@ static unsigned char *fb_init_driver(unsigned char *param, unsigned char *ignore
 					data[i] = val++;
 				memcpy(from, data + start, end - start);
 				/*fprintf(stderr, "copy %lu: %p, %p, %x\n", iteration, test + start, from, end - start);*/
-				memcpy_to_fb(test + start, from, end - start);
+				memcpy_to_fb(test + start, from, end - start, 1);
 				if (memcmp(test, data, LEN))
 					break;
 			}
 			iteration++;
 		}
-		internal("memcpy is broken (%lu): %u, %u (%u), %p", iteration, start, end, end - start, from);
+		internal_error("memcpy is broken (%lu): %u, %u (%u), %p", iteration, start, end, end - start, from);
 	}
 #undef LEN
 #endif
@@ -1171,7 +1308,8 @@ static unsigned char *fb_init_driver(unsigned char *param, unsigned char *ignore
 			e = stracpy(cast_uchar "-mode syntax is left_border[,top_border[,right_border[,bottom_border]]]\n");
 			goto fail1;
 		}
-		ul = strtoul(cast_const_char param, (char **)(void *)&param, 10);
+		ul = strtoul(cast_const_char param, &se, 10);
+		param = cast_uchar se;
 		if (ul > MAXINT / 10) goto bad_p;
 		border_left = (int)ul;
 		if (*param == ',') param++;
@@ -1180,7 +1318,8 @@ static unsigned char *fb_init_driver(unsigned char *param, unsigned char *ignore
 	}
 	if (*param) {
 		if (*param < '0' || *param > '9') goto bad_p;
-		ul = strtoul(cast_const_char param, (char **)(void *)&param, 10);
+		ul = strtoul(cast_const_char param, &se, 10);
+		param = cast_uchar se;
 		if (ul > MAXINT / 10) goto bad_p;
 		border_top = (int)ul;
 		if (*param == ',') param++;
@@ -1189,7 +1328,8 @@ static unsigned char *fb_init_driver(unsigned char *param, unsigned char *ignore
 	}
 	if (*param) {
 		if (*param < '0' || *param > '9') goto bad_p;
-		ul = strtoul(cast_const_char param, (char **)(void *)&param, 10);
+		ul = strtoul(cast_const_char param, &se, 10);
+		param = cast_uchar se;
 		if (ul > MAXINT / 10) goto bad_p;
 		border_right = (int)ul;
 		if (*param == ',') param++;
@@ -1198,7 +1338,8 @@ static unsigned char *fb_init_driver(unsigned char *param, unsigned char *ignore
 	}
 	if (*param) {
 		if (*param < '0' || *param > '9') goto bad_p;
-		ul = strtoul(cast_const_char param, (char **)(void *)&param, 10);
+		ul = strtoul(cast_const_char param, &se, 10);
+		param = cast_uchar se;
 		if (ul > MAXINT / 10) goto bad_p;
 		border_bottom = (int)ul;
 		if (*param == ',') param++;
@@ -1428,10 +1569,7 @@ retry1:
 	virtual_devices[0] = NULL;
 	global_mouse_hidden = 1;
 	last_mouse_buttons = B_MOVE;
-	if (handle_fb_mouse()) {
-		e = stracpy(cast_uchar "Cannot open GPM mouse.\n");
-		goto fail7;
-	}
+	handle_fb_mouse();
 	/* hide cursor */
 	if (border_left | border_top | border_right | border_bottom) fb_clear_videoram();
 
@@ -1443,10 +1581,11 @@ retry1:
 	if (getenv("LINKS_MEMCPY_BENCH")) {
 		unsigned array2[20000];
 		unsigned i;
-		for (i = 0; i < 32768000; i++) {
+		for (i = 0; i < 327680; i++) {
 			/*memcpy_to_fb((unsigned char *)fb_mem, (unsigned char *)array2 + 0, 2048);*/
 			/*memcpy_to_fb((unsigned char *)fb_mem + (i & 31), (unsigned char *)array2 + ((i >> 5) & 31), 11000);*/
-			memcpy_to_fb((unsigned char *)fb_mem + (i & 31), (unsigned char *)array2 + ((i >> 5) & 31), ((i >> 10) & 31) + 1);
+			/*memcpy_to_fb((unsigned char *)fb_mem + (i & 31), (unsigned char *)array2 + ((i >> 5) & 31), ((i >> 10) & 31) + 1, 1);*/
+			memcpy_to_fb((unsigned char *)array2 + ((i >> 5) & 31), (unsigned char *)fb_mem + (i & 31), ((i >> 10) & 8191) + 1, 1);
 		}
 		exit(1);
 	}
@@ -1454,12 +1593,14 @@ retry1:
 
 	return NULL;
 
+#if 0
 fail7:
 	fb_driver.shutdown_device(mouse_graphics_device);
 	mem_free(mouse_buffer);
 	mem_free(background_buffer);
 	mem_free(new_background_buffer);
 	fb_shutdown_palette();
+#endif
 fail6:
 	EINTRLOOP(rs, munmap(fb_mem, fb_mapped_size));
 fail5:
@@ -1502,7 +1643,7 @@ static void fb_shutdown_driver(void)
 	fb_show_cursor();
 	install_signal_handler(SIGINT, (void (*)(void *))NULL, NULL, 0);
 
-	if (in_gr_operation) internal("fb_shutdown_driver: in_gr_operation %d", in_gr_operation);
+	if (in_gr_operation) internal_error("fb_shutdown_driver: in_gr_operation %d", in_gr_operation);
 }
 
 static void fb_after_fork(void)
@@ -1593,10 +1734,15 @@ static int fb_set_margin(int left, int right, int top, int bottom)
  */
 static int fb_get_empty_bitmap(struct bitmap *dest)
 {
-	if (dest->x && (unsigned)dest->x * (unsigned)dest->y / (unsigned)dest->x != (unsigned)dest->y) overalloc();
-	if ((unsigned)dest->x * (unsigned)dest->y > (unsigned)MAXINT / fb_pixelsize) overalloc();
-	dest->data = mem_alloc(dest->x * dest->y * fb_pixelsize);
-	dest->skip = dest->x * fb_pixelsize;
+	dest->data = NULL;
+	if (dest->x && (size_t)dest->x * (size_t)dest->y / (size_t)dest->x != (size_t)dest->y)
+		return -1;
+	if ((size_t)dest->x * (size_t)dest->y > MAX_SIZE_T / 2 / fb_pixelsize)
+		return -1;
+	dest->data = mem_alloc_mayfail((size_t)dest->x * (size_t)dest->y * fb_pixelsize);
+	if (!dest->data)
+		return -1;
+	dest->skip = (ssize_t)dest->x * fb_pixelsize;
 	dest->flags = 0;
 	return 0;
 }
@@ -1607,11 +1753,14 @@ static void fb_register_bitmap(struct bitmap *bmp)
 
 static void fb_unregister_bitmap(struct bitmap *bmp)
 {
-	mem_free(bmp->data);
+	if (bmp->data)
+		mem_free(bmp->data);
 }
 
 static void *fb_prepare_strip(struct bitmap *bmp, int top, int lines)
 {
+	if (!bmp->data)
+		return NULL;
 	return ((unsigned char *)bmp->data) + bmp->skip * top;
 }
 
@@ -1634,7 +1783,7 @@ static void fb_draw_bitmap(struct graphics_device *dev, struct bitmap *bmp, int
 	scr_start = fb_vmem + y * fb_linesize + x * fb_pixelsize;
 	copy_len = xs * fb_pixelsize;
 	for (; ys; ys--) {
-		memcpy_to_fb_inline(scr_start, data, copy_len);
+		memcpy_to_fb_inline(scr_start, data, copy_len, 0);
 		data += bmp->skip;
 		scr_start += fb_linesize;
 	}
@@ -1789,7 +1938,7 @@ static void fb_draw_vline(struct graphics_device *dev, int x, int y1, int y2, lo
 #endif
 			default: {
 				do {
-					memcpy_to_fb(dest, (unsigned char *)&color, fb_pixelsize);
+					memcpy_to_fb(dest, (unsigned char *)&color, fb_pixelsize, 0);
 					dest += fb_linesize;
 				} while (--y);
 				break;
@@ -1860,7 +2009,7 @@ static int fb_scroll(struct graphics_device *dev, struct rect_set **ignore, int
 			dest += dest_off;
 			src += src_off;
 			for (y = dev->clip.y2 - dev->clip.y1 - scy; y; y--) {
-				memcpy(dest, src, len);
+				memcpy_to_fb(dest, src, len, 1);
 				dest -= fb_linesize;
 				src -= fb_linesize;
 			}
@@ -1872,9 +2021,9 @@ static int fb_scroll(struct graphics_device *dev, struct rect_set **ignore, int
 			src += src_off;
 			for (y = dev->clip.y2 - dev->clip.y1 + scy; y; y--) {
 				if (scy)
-					memcpy(dest, src, len);
+					memcpy_to_fb(dest, src, len, 1);
 				else
-					memmove(dest, src, len);
+					memmove_in_fb(dest, src, len);
 				dest += fb_linesize;
 				src += fb_linesize;
 			}
@@ -1895,7 +2044,7 @@ static void fb_update_palette(void)
 		fb_driver.depth &= ~0x300;
 	fb_driver.get_color = get_color_fn(fb_driver.depth);
 	if (!fb_driver.get_color)
-		internal("invalid driver depth %d", fb_driver.depth);
+		internal_error("invalid driver depth %d", fb_driver.depth);
 
 	mouse_black = fb_driver.get_color(0);
 	mouse_white = fb_driver.get_color(0xffffff);
@@ -1943,9 +2092,7 @@ static int fb_unblock(struct graphics_device *dev)
 	current_virtual_device = fb_old_vd;
 	fb_old_vd = NULL;
 	if (have_cmap) set_palette(&global_pal);
-	if (handle_fb_mouse()) {
-		fatal_exit("Cannot open GPM mouse.");
-	}
+	handle_fb_mouse();
 	END_GR
 	if (border_left | border_top | border_right | border_bottom) fb_clear_videoram();
 	if (current_virtual_device) current_virtual_device->redraw_handler(current_virtual_device, &current_virtual_device->size);
@@ -1960,7 +2107,7 @@ struct graphics_driver fb_driver = {
 	init_virtual_device,
 	shutdown_virtual_device,
 	fb_shutdown_driver,
-	(void (*)(void))NULL,
+	fb_show_cursor,
 	fb_after_fork,
 	fb_get_driver_param,
 	(unsigned char *(*)(void))NULL,
@@ -1982,6 +2129,7 @@ struct graphics_driver fb_driver = {
 	fb_block,
 	fb_unblock,
 	fb_update_palette,						/* set_palette */
+	NULL,								/* get_real_colors */
 	(void (*)(struct graphics_device *, unsigned char *))NULL,	/* set_title */
 	(int (*)(unsigned char *, int))NULL,				/* exec */
 	(void (*)(struct graphics_device *, unsigned char *))NULL,	/* set_clipboard_text */
