@@ -987,7 +987,6 @@ void close_fork_tty(void)
 	struct k_conn *k;
 	struct list_head *lk;
 	int rs;
-	reinit_child(1);
 #ifndef NO_SIGNAL_HANDLERS
 	EINTRLOOP(rs, close(signal_pipe[0]));
 #endif
@@ -1336,7 +1335,7 @@ int exe(unsigned char *path, int fg)
 	if (p > 0) {
 		EINTRLOOP(rp, waitpid(p, &s, 0));
 	} else {
-		system(path);
+		rs = system(path);
 		return rs;
 	}
 	return 0;
@@ -3181,6 +3180,8 @@ int os_receive_fg_cookie(int h)
 }
 #endif
 
+int need_detach_console = 0;
+
 void os_detach_console(void)
 {
 #if defined(WIN)
@@ -3196,7 +3197,7 @@ void os_detach_console(void)
 #endif
 		EINTRLOOP(rp, fork());
 		if (!rp) {
-			reinit_child(0);
+			reinit_child();
 #if defined(HAVE_PTHREADS)
 			reset_thread_count();
 #endif
